#! /bin/sh
# Script to upgrade the OS of a board
#
# Upgrade one of the two sets (1 or 2) of OS on a board

set -e

MAGIC_OS1=1330390856
MAGIC_OS2=1265782898
DEVENVSIZE=0x1000

# dryrun_echo="echo"
dryrun_echo=

# Default names:
path_kernel="uImage"
path_rootfs="core-image-minimal-dev-vp-vpe-zynq7.ubifs"

# Set _temporary_ paths for the DTB and FPGA, they will be reset below
# when we will know the product:
product="<product>"
path_dtb="${product}.dtb.lzo.u-boot"
path_fpga="${product}.bit.lzo.u-boot"

usage="upgrd_os.sh <product> <osnbr> [-c <seqnbrfile>] -p <package>
    <product>:     product name
    <osnbr>:       OS set number (1 | 2 | 0 = auto)
    <seqnbrfile>:  path to a NAND OS-set sequence number file
    <package>:     path to a VP Upgrade Package file
"

# Check if product matches board name:
if [ "X$1" == "X" ] ; then
    echo "$usage"
    echo "Error: arg 1 must be the product name"
    exit 10
fi
product="$1"
if [ "$product" != `cat /proc/device-tree/vp_boot/product` ] ; then
    echo "Error: product don't match"
    exit 12
fi

# Set DTB and FPGA config names for the product:
path_dtb="${product}.dtb.lzo.u-boot"
path_fpga="${product}.bit.lzo.u-boot"

# Check if board features xc7z035 (before rev 4.0) or xc7z045 (from rev 4.0) FPGA
vphw_xc7z045_rev=4.0
vphw=`cat /proc/device-tree/vp_boot/vphw`
vphw_ver=${vphw##*-v}
if [[ ${vphw_ver%.*} -ge ${vphw_xc7z045_rev%.*} ]] ; then
   path_fpga="${product}-rev4.bit.lzo.u-boot"
fi

# Check if OS set 1 or set 2, set the MTD numbers accordingly:
if [ "x$2" == "x" ] ; then
    echo "$usage"
    echo "Error: arg 1 is the os set number, must be 0 (auto-detect), 1 or 2"
    exit 11
elif [ "x$2" == "x0" ] ; then
    rootfs=`grep rootfs /proc/mounts | cut -d \  -f 1`
    if [ "$rootfs" == "ubi0:rootfs2" ] ; then
        osnbr="1"
    elif [ "$rootfs" == "ubi0:rootfs1" ] ; then
        osnbr="2"
    else
        echo "Error: unable to determine current OS number "
        exit 11
    fi
else
    osnbr="$2"
fi

if [ "$osnbr" == "1" ] ; then
    mtd_kernel="3"
    mtd_dtb="4"
    mtd_fpga="5"
    ubi_id="0"
    current_osnbr="2"
    mtd_current_kernel="6"
elif [ "$osnbr" == "2" ] ; then
    mtd_kernel="6"
    mtd_dtb="7"
    mtd_fpga="8"
    ubi_id="1"
    current_osnbr="1"
    mtd_current_kernel="3"
else
    echo "$usage"
    echo "Error: arg 1 is the os set number, must be 0 (auto-detect), 1 or 2"
    exit 11
fi

# Set u-boot SPL path and MTD number:
path_spl="boot.bin"
mtd_spl="0"

path_uboot="u-boot-dtb.img"
mtd_uboot="1"

# Check if we should flash an OS-set sequence number file:
path_seqnbr_file=""
if [ "x$3" == "x-c" ] ; then
    if [ "x$4" == "x" ] ; then
        echo "$usage"
        echo "Error: when '-c' is specified, the <seqnbrfile> must be specified too"
        exit 31
    fi
    path_seqnbr_file="$4"
    # Check the presence of the os sequence number file:
    if [ ! -f "${path_seqnbr_file}" ] ; then
        echo "$usage"
        echo "Error: os sequence number file '${path_seqnbr_file}' does not exist."
        exit 32
    fi
    shift 2
fi

function read_seqnbr ()
{
    size="0x`grep mtd${mtd_current_kernel}: /proc/mtd | cut -d \  -f 2`"
    erase_size="0x`grep mtd${mtd_current_kernel}: /proc/mtd | cut -d \  -f 3`"

    # The environment section should not exceed DEVENVSIZE, which we read in 512 byte blocks
    read_size=$(( ${DEVENVSIZE} / 512 ))
    skip=$(( ( $size - $erase_size  - $erase_size ) / 512 ))
    dd if=/dev/mtd${mtd_current_kernel} of=/tmp/currenv.txt count=${read_size} skip=${skip}
    if [ $? -ne 0 ]
    then
        # Assume bad block, try the next block
        skip=$(( $skip + ( $erase_size / 512 ) ))
        dd if=/dev/mtd${mtd_current_kernel} of=/tmp/currenv.txt count=${read_size} skip=${skip}
        if [ $? -ne 0 ]
        then
            # We've completely failed to read the sequence number. How did we
            # even manage to boot this bank?
            echo 1
            return
        fi
    fi

    # This errors, because there's -xff at the end. Ignore the error.
    source /tmp/currenv.txt 2>/dev/null || true

    if [ "${current_osnbr}" == "1" ]
    then
        if [ "${os1magic}" == "${MAGIC_OS1}" ]
        then
            echo ${os1seqnbr}
            return
        fi
    else
        if [ "${os2magic}" == "${MAGIC_OS2}" ]
        then
            echo ${os2seqnbr}
            return
        fi
    fi

    # We can't read our own sequence number. Assume this means there is no
    # sequence number, and start from 1.
    echo 1
    return
}

function make_seqnbr_file ()
{
    osnbr=$1
    seqnbr=$2
    file=$3

    echo "os${osnbr}seqnbr=${seqnbr}" > ${file}

    if [ "x${osnbr}" == "x1" ]
    then
        echo os1magic=${MAGIC_OS1} >> ${file}
    else
        echo os2magic=${MAGIC_OS2} >> ${file}
    fi

    # Ensure it's terminated by a 0 byte
    printf \\0 >> ${file}
}

function write_seqnbr ()
{
    file=$1
    size="0x`grep mtd${mtd_kernel}: /proc/mtd | cut -d \  -f 2`"
    erase_size="0x`grep mtd${mtd_kernel}: /proc/mtd | cut -d \  -f 3`"

    # Start write on the second to last block of the mtd partition
    # If that's bad nandwrite will automatically try on the last block
    offset=$(( $size - $erase_size - $erase_size ))
    ${dryrun_echo} nandwrite -m -p -s ${offset} /dev/mtd${mtd_kernel} ${file}
    if [ $? -ne 0 ]
    then
        # We're in trouble. Neither block is usable.
        logger VP upgrade: Failed to write sequence number: nandwrite returned $?
        exit 1
    fi
}

if [ -z ${path_seqnbr_file} ]
then
    seqnbr=`read_seqnbr ${other_osnbr}`
    newseqnbr=$(($seqnbr + 1))

    path_seqnbr_file=/tmp/seqnbr.txt
    make_seqnbr_file $osnbr $newseqnbr ${path_seqnbr_file}
fi

# Check if we use an upgrade package:
if [ "x$3" == "x-p" ] ; then
    if [ "x$4" == "x" ] ; then
        echo "$usage"
        echo "Error: when '-p' is specified, the <package> must be specified too"
        exit 21
    fi
    path_package="$4"
    # Check the presence of the upgrade package file:
    if [ ! -f "${path_package}" ] ; then
        echo "$usage"
        echo "Error: package file '${path_package}' does not exist."
        exit 22
    fi
else
	echo "$usage"
	exit 23
fi

function flash_an_mtd ()
{
    mtdnbr="$1"
    filepath="$2"
    ${dryrun_echo} flash_erase /dev/mtd${mtdnbr} 0 0
    ${dryrun_echo} nandwrite -m -p /dev/mtd${mtdnbr} ${filepath}
}

function flash_uboot_spl ()
{
    filepath="$1"

    ${dryrun_echo} nanddump -s 126976 -l 2048 -f /tmp/mtd0.bin /dev/mtd${mtd_spl}
    ${dryrun_echo} flash_erase /dev/mtd${mtd_spl} 0 0
    ${dryrun_echo} nandwrite -m -p /dev/mtd${mtd_spl} ${filepath}
    ${dryrun_echo} nandwrite -s 126976 -m -p /dev/mtd${mtd_spl} /tmp/mtd0.bin
}

function untar()
{
	filename=$1

	${dryrun_echo} tar --wildcards --numeric-owner --warning=no-timestamp --to-stdout -xzf "${path_package}" ${filename}
}

function check_file()
{
	filename=$1
	${dryrun_echo} tar --wildcards --numeric-owner --warning=no-timestamp -tf "${path_package}" | grep ${filename}
}

function find_mtd()
{
	cat /proc/mtd | awk "BEGIN { FS=\":\" } /$1/ { print \"/dev/\" \$1; }"
}

# Step zero: Verify that we have the FPGA/DTB images for this board
if ! check_file ${path_dtb}
then
	logger VPUP does not contain ${path_dtb}
	exit 30
fi
if ! check_file ${path_fpga}
then
	logger VPUP does not contain ${path_fpga}
	exit 31
fi

# Erase the extra environment so we don't risk having a version which does not
# support alternate FPGA images boot with an alternate FPGA.
mtd_extra_env=`find_mtd nand-extra-env`
if [ -n "${mtd_extra_env}" ]
then
	logger Erasing extra environment on MTD ${mtd_extra_env}
	flash_eraseall ${mtd_extra_env}
fi

# First step: Erase the OS sequential number file
write_seqnbr /dev/zero

untar ${path_kernel} | flash_an_mtd ${mtd_kernel} -
untar ${path_dtb}    | flash_an_mtd ${mtd_dtb} -
untar ${path_fpga}   | flash_an_mtd ${mtd_fpga} -

if tar --numeric-owner --warning=no-timestamp -xzf "${path_package}" ${path_spl} >/dev/null 2>&1; then
    flash_uboot_spl ${path_spl}
fi

if tar --numeric-owner --warning=no-timestamp -xzf "${path_package}" ${path_uboot} >/dev/null 2>&1; then
    flash_an_mtd ${mtd_uboot} ${path_uboot}
fi

# More complex case: rootfs still in the package, we do not want
# to unpack it because it is too big, so we pipe the rootfs from
# "tar"'s stdout to "ubiupdatevol"'s stdin. But in this case, we
# must specify the input size.
# 1. Fetch the rootfs size by parsing the output of "tar -t":
size=`tar --wildcards --numeric-owner -tvzf "${path_package}" '*.ubifs' | { read perms user size rest ; echo "${size}" ; }`
# 2. Extract from "tar" to stdout and pipe in stdin of
#    "ubiupdatevol", using the "--size=" option:
${dryrun_echo} tar --wildcards --numeric-owner --warning=no-timestamp --to-stdout -xzf "${path_package}" '*.ubifs' | ${dryrun_echo} ubiupdatevol --size=${size} /dev/ubi0_${ubi_id} -

write_seqnbr ${path_seqnbr_file}

exit 0
